from dataclasses import dataclass
from typing import Callable, List, Optional
from lsfb_dataset.datasets.types import *
import os
import pandas as pd


@dataclass
class LSFBIsolConfig:
    """
    Simple configuration class for the lsfb_isol Dataset.

    lsfb: French Belgian Sign Language
    isol: isolated videos in LSFB

    Args:
        root: Root directory of the LSFB_ISOL dataset.
            The dataset must already be downloaded.

        landmarks: Select which landmarks (features) to use. Default = ['pose', 'hand_left', 'hand_right'].
            'pose' for pose skeleton (23 landmarks);
            'hands_left' for left hand skeleton (21 landmarks);
            'hands_right' for right hand skeleton (21 landmarks);

        features_transform: Callable object used to transform the features.
        target_transform: Callable object used to transform the targets.
        transform: Callable object used to transform both the features and the targets.

        mask_transform: Callable object used to transform the masks.
            You need to set return_mask to true to use this transform.

        lemmes_nb: Number of lemme to consider. Default=10
        lemme_list_file: Path to the csv containing the lemmes lists. Default="lemmes.csv"
        videos_list_file: Path to the csv containing the video information. Default="clips.csv"

        split: Select a specific subset of the dataset. Default = 'all'.
            'train' for training set;
            'test' for the test set;
            'all' for all the instances of the dataset;
            'mini_sample' for a tiny set of instances.

        sequence_max_length: Max length of the clip sequence. Default=50.
        padding: Pad all sequence to the same length.
        return_mask: Returning padding mask for the sequence.
        mask_value: Value of the masked part of the clips.
        show_progress: If true, show a progress bar while the dataset is loading.


    Author: jfink
    """

    root: str
    landmarks: Optional[List[str]] = None

    features_transform: Callable = None
    target_transform: Callable = None
    transform: Callable = None
    mask_transform: Callable = None

    lemmes_nb: int = 10
    lemme_list_file: str = 'lemmes.csv'
    videos_list_file: str = 'clips.csv'

    split: DataSubset = 'all'
    sequence_max_length: int = 50

    padding: bool = False
    return_mask: bool = False
    mask_value: int = 0
    show_progress: bool = True

    def __post_init__(self):
        if self.landmarks is None:
            self.landmarks = ['pose', 'hand_left', 'hand_right']

        self.lemme_list_path = os.path.join(self.root, self.lemme_list_file)
        self.videos_list_path = os.path.join(self.root, self.videos_list_file)

        self.videos = pd.read_csv(self.videos_list_path)

        self.lemmes = pd.read_csv(self.lemme_list_path)
        self.lemmes = self.lemmes.iloc[:self.lemmes_nb]



"""
## lsfb_isol

This module contains classes helping to manipulate the lsfb_isol dataset. The provided classes are :

- **LSFBIsolConfig** : [dataclass](https://docs.python.org/3/library/dataclasses.html) allowing to configure
how the dataset will be loaded
- **LSFBIsolLandmarks** : Iterator class loading the lsfb_isol landmarks according to the configuration provided.
- **LSFBIsolLandmarksGenerator** : Generator class loading the lsfb_isol landmarks according to the configuration
provided.

"""

from .landmarks import LSFBIsolLandmarks
from .landmarks_generator import LSFBIsolLandmarksGenerator
from .config import LSFBIsolConfig



import cv2
from ..utils.cv import absolute_position


def draw_line(img, start_pos, end_pos, color=(255, 255, 255), thickness=2):
    cv2.line(img, absolute_position(img, start_pos), absolute_position(img, end_pos), color, thickness)


def draw_connections(img, positions, connections, color=(255, 0, 0), thickness=2):
    pos_nb = len(positions)
    for i, j in connections:
        if i < pos_nb and j < pos_nb:
            x1, y1 = absolute_position(img, positions[i])
            x2, y2 = absolute_position(img, positions[j])
            cv2.line(img, (x1, y1), (x2, y2), color, thickness)


def draw_points(img, positions, color=(0, 0, 255), thickness=cv2.FILLED, radius=3):
    for pos in positions:
        x, y = absolute_position(img, pos)
        cv2.circle(img, (x, y), radius, color, thickness)


def draw_rect(img, pos1, pos2, color=(255, 255, 255), thickness=cv2.FILLED):
    cv2.rectangle(img, absolute_position(img, pos1), absolute_position(img, pos2), color=color, thickness=thickness)






from urllib.parse import quote, urljoin
from pathlib import PurePath
import requests
import hashlib
import pandas as pd
from tqdm import tqdm
import os
from os import path
from typing import List
from ..datasets import split_cont, split_isol, mini_sample


class DatasetDownloader:
    """
    DatasetDownloader 

    This class provides an easy interface for downloading the LSFB dataset over http.

    Args:
        destination : Path to the destination folder where the dataset is saved.
        dataset : The dataset to download. Default = 'isol'.
            'isol' for the LSFB isol dataset;
            'cont' for the LSFB cont dataset;
        landmarks : Select which landmarks (features) to download. Default = ['pose', 'hands'].
            'pose' for pose skeleton (23 landmarks);
            'hands' for hand skeleton (21 landmarks per hand);
            'face' for face skeleton (468 landmarks).
        split : Select which subset of the dataset should be downloaded. Default = 'all'.
            'all' for the entire dataset;
            'train' for the training data only;
            'test' for the test data only;
            'mini_sample' for a tiny sample of the dataset.
        include_video : Download the raw video. Default = False.
        include_raw : Download the raw landmarks (features) without smoothing nor interpolation.
            Default = False.
        compute_hash : Verify the hash of the video. Include_video should be set to true to use
            this options. Default = False.
        src : URL of the source of the dataset. Default = URL of the UNamur servers
    """

    def __init__(
            self,
            destination: str,
            dataset: str = "isol",
            landmarks: List[str] = None,
            split: str = 'all',
            include_video: bool = False,
            include_raw: bool = False,
            compute_hash: bool = False,
            src: str = None,
    ):
        self.destination = destination
        self.dataset = dataset
        self.split = split

        if landmarks is None:
            self.landmarks = ["pose", "hands"]
        else:
            self.landmarks = landmarks

        if include_raw:
            landmarks = [*self.landmarks]
            for landmark in landmarks:
                self.landmarks.append(landmark + "_raw")

        self.include_video = include_video
        self.compute_hash = compute_hash

        if src is not None:
            self.src = src
        elif dataset == "isol":
            self.src = "https://lsfb.info.unamur.be/static/datasets/LSFB/LSFB_ISOL"
        elif dataset == "cont":
            self.src = "https://lsfb.info.unamur.be/static/datasets/LSFB/LSFB_CONT"

    def download(self):
        """
        Download the data accordingly to the downloader configuration.
        """
        csv_path = self._download_csv()
        data = pd.read_csv(csv_path)
        data = self._select_split(data)

        progress_bar = tqdm(data.iterrows(), total=data.shape[0])
        for idx, row in progress_bar:
            progress_bar.set_postfix_str(row['filename'])

            if self.include_video:
                self.download_video(row)

            if self.dataset == "cont":
                self._download_annotations(row, "annots_left")
                self._download_annotations(row, "annots_right")
                self._download_annotations(row, "annots_trad")
                self._download_segmentation_vectors()

            for landmark in self.landmarks:
                if type(row[landmark]) == str and row[landmark] != "":
                    self._download_landmarks(row[landmark])

    def _download_csv(self):
        csv_files = []

        if self.dataset == "isol":
            csv_files.append("lemmes.csv")
            csv_files.append("clips.csv")
            metadata_dest = path.join(self.destination, "clips.csv")

        elif self.dataset == "cont":
            csv_files.append("videos.csv")
            metadata_dest = path.join(self.destination, "videos.csv")

        else:
            raise ValueError(f'Unknown dataset: {self.dataset}')

        for origin_path in csv_files:
            destination = os.path.join(self.destination, origin_path)
            self._download_file(origin_path, destination)

        return metadata_dest

    def download_video(self, row):
        """
        Download the video from the source url if the video was not yet downloaded by the client.
        This function also checks the integrity of the video already downloaded. If the md5 sum doesn't match
        the video is re-downloaded.

        Input:

        row : The dataframe row containing the information for that video.
        """

        if self.dataset == 'isol':
            video_path = row["video"]
        else:
            video_path = row["filepath"]

        video_destination = os.path.join(self.destination, video_path)

        if not os.path.isfile(video_destination):
            self._download_file(video_path, video_destination)

        elif os.path.getsize(video_destination) == 0:
            self._download_file(video_path, video_destination)

        elif self.dataset == "cont" and self.compute_hash:
            # Check md5sum
            dest_md5 = hashlib.md5(open(video_destination, "rb").read()).hexdigest()
            src_md5 = row["md5"]

            if dest_md5 != src_md5:
                self._download_file(video_path, video_destination)

    def _download_annotations(self, row, col_name):
        """
        Download the annotations file for a video.
        There is one file for each hand.

        Input:
        row : The dataframe row containing the information for that video.
        hand : The hand you want to download the annotations for. Either "right_hand" or "left_hand".
        """

        annot_path = row[col_name]
        annotation_destination = os.path.join(self.destination, annot_path)
        if not os.path.exists(annotation_destination):
            self._download_file(annot_path, annotation_destination)

    def _download_segmentation_vectors(self):
        destinations = [
            (f'annotations/vectors/{x}.pck', path.join(self.destination, 'annotations/vectors', f'{x}.pck'))
            for x in ['activity', 'binary', 'binary_with_coarticulation']
        ]
        for origin, dest in destinations:
            self._download_file(origin, dest)

    def _download_landmarks(self, landmark_path):
        landmarks_destination = os.path.join(self.destination, landmark_path)
        if not os.path.exists(landmarks_destination):
            self._download_file(landmark_path, landmarks_destination)

    def _download_file(self, origin: str, destination: str):
        url = urljoin(f'{self.src}/', quote(origin))

        req = requests.get(url, allow_redirects=False, timeout=60)
        if req.status_code != 200:
            raise FileNotFoundError(f"""
            Could not download the desired file: {origin}.
            Wrong  HTTP status ({req.status_code}) for request ({url}).
            """)

        os.makedirs(str(PurePath(destination).parent), exist_ok=True)
        with open(destination, 'wb') as file:
            file.write(req.content)

    def _select_split(self, data):
        if self.split == 'all':
            return data

        if self.split == 'mini_sample':
            return mini_sample(data)

        if self.dataset == 'isol':
            train, test = split_isol(data)
        else:
            train, test = split_cont(data)

        if self.split == 'train':
            return train
        elif self.split == 'test':
            return test
        else:
            raise ValueError(f'Unknown split: {self.split}')



from .base import LSFBContBase
from ...utils.datasets import create_mask
from ...utils.landmarks import load_landmarks, pad_landmarks
from ...utils.target import pad_target


class LSFBContLandmarksGenerator(LSFBContBase):
    """
    Utility class to load the LSFB CONT Landmarks dataset.
    The dataset must be already downloaded!

    All the landmarks and targets are lazy loaded. In consequence, iterate over all the instances is slower
    but consumes less memory (RAM).

    If you have enough RAM (more than 16GB) and want more efficient iterations,
    use the LSFBContLandmarks class instead.

    Properties:
        targets: The list of target segmentations for each instance of the dataset

        labels: The targeted labels in the dataset. Example: waiting, signing and coarticulation.
            The labels depend on the targeted segmentation.
        label_frequencies: The frequency of each label used in the dataset.
            The labels depend on the targeted segmentation.

        windows: List of every window in the dataset
            (instance_index, window_start, window_end, padding)
            This list only contains windows if the `window` configuration is set!

    Args:
        config: The configuration object (see LSFBContConfig).
            If config is not specified, every needed configuration argument must be manually provided.

    Author: ppoitier
    """

    def __init__(self, **kwargs):
        super(LSFBContLandmarksGenerator, self).__init__(**kwargs)

        # (video_index, window_start, window_end, padding)
        self.windows: list[tuple[int, int, int, int]] = []
        if self.config.window is not None:
            self._build_windows()

    def __len__(self):
        if self.config.window is None:
            return self.videos.shape[0]
        return len(self.windows)

    def __getitem__(self, index):
        features_transform = self.config.features_transform
        target_transform = self.config.target_transform
        transform = self.config.transform

        window = self.config.window
        return_mask = self.config.return_mask
        mask_transform = self.config.mask_transform

        padding = 0

        if window is not None:
            features, target, padding = self._get_windowed_item(index)
        else:
            features = load_landmarks(
                self.videos.iloc[index],
                self.config.root,
                self.config.landmarks
            )
            target = self.targets[index]

        if features_transform is not None:
            features = features_transform(features)

        if target_transform is not None:
            target = target_transform(target)

        if transform is not None:
            features, target = transform(features, transform)

        if return_mask:
            mask = create_mask(len(target), padding)
            if mask_transform is not None:
                mask = mask_transform(mask)
            return features, target, mask

        return features, target

    def _get_windowed_item(self, index):
        video_index, start, end, padding = self.windows[index]
        features = load_landmarks(
            self.videos.iloc[video_index],
            self.config.root,
            self.config.landmarks
        )
        target = self.targets[video_index]

        landmarks = pad_landmarks(features[start:end], padding)
        target = pad_target(target[start:end], padding)
        return landmarks, target, padding

    def _build_windows(self):
        window = self.config.window
        window_size, window_stride = window
        for index, seq_len in enumerate(self.videos['frames']):
            for start in range(0, seq_len, window_stride):
                end = min(seq_len, start + window_size)
                padding = window_size - (end - start)
                self.windows.append((index, start, end, padding))
        print('Windows successfully created.')



from dataclasses import dataclass
from os import path
from typing import List, Optional, Tuple, Callable
from ..types import *


@dataclass
class LSFBContConfig:
    """
    Simple configuration class for the LSFB_CONT Dataset.

    LSFB: French Belgian Sign Language
    CONT: Continuous videos in LSFB

    See: LSFBContLandmarks

    Args:
        root: Root directory of the LSFB_CONT dataset.
            The dataset must already be downloaded !
        landmarks: Select which landmarks (features) to use.
            'pose' for pose skeleton (23 landmarks);
            'hand_left' for left hand skeleton (21 landmarks);
            'hand_right' for right hand skeleton (21 landmarks);
            Default = ['pose', 'hand_left', 'hand_right'].

        features_transform: Callable object used to transform the features.
        target_transform: Callable object used to transform the targets.
        transform: Callable object used to transform both the features and the targets.

        split: Specify which subset of the dataset is used.
            'train' for training set;
            'test' for the test set;
            'all' for all the instances of the dataset;
            'mini_sample' for a tiny set of instances.
            Default = 'all'.
        hands: Specify which hands are targeted in the segmentation.
            'right' for the signs from the right hand of the signer;
            'left' for the signs from the left hand of the signer;
            'both' for the signs from both hands of the signer.
            Default = 'both'.
        target: Specify the kind of segmentation used as a target.
            'signs' for the labels waiting and signing only.
            'signs_and_transitions' for the labels waiting, signing and coarticulation.
            'activity' for the labels waiting and signing,
                where intermediate movements are labelled as signing and not waiting anymore.
            Default = 'signs'.

        window: Optional argument (window_size, window_stride) to use fixed-size windows instead of
            variable length sequences.
            If specified, the dataset is windowed with a window size and a window stride.
            Be careful, this argument changes the number of instances in the dataset !
            Default = None.
        return_mask: If true, return a mask in addition to the features and the target for each instance.
            The mask is filled with 1's and 0's where the padding has been applied.
            Default = False.
        mask_transform: Callable object used to transform the masks.
            You need to set return_mask to `True` to use this transform !

        video_list_file: The filepath of the video list CSV file.
            This filepath is relative to the root of the dataset.
        targets_dir: The path of the directory that contains the target vectors.
            This filepath is relative to the root of the dataset.

        show_progress: If true, shows a progress bar while the dataset is loading.

    Author: ppoitier
    """

    root: str
    landmarks: Optional[List[str]] = None

    features_transform: Callable = None
    target_transform: Callable = None
    transform: Callable = None

    split: DataSubset = 'all'
    hands: Hand = 'both'
    target: Target = 'signs'

    window: Optional[Tuple[int, int]] = None
    return_mask: bool = False
    mask_transform: Callable = None

    video_list_file: str = 'videos.csv'
    targets_dir: str = 'annotations/vectors'
    show_progress: bool = True

    def __post_init__(self):
        if self.landmarks is None:
            self.landmarks = ['pose', 'hand_left', 'hand_right']

        self.video_list_file = path.join(self.root, self.video_list_file)
        self.targets_dir = path.join(self.root, self.targets_dir)

        self.validate()

    def validate(self):
        if self.mask_transform is not None and self.return_mask is False:
            raise ValueError('Mask transform is defined, but the mask is not returned.')



import pandas as pd
from tqdm import tqdm
from datetime import datetime

from typing import List
from lsfb_dataset.utils.landmarks import load_pose_landmarks, load_hands_landmarks, pad_landmarks
from lsfb_dataset.utils.datasets import create_mask
from lsfb_dataset.datasets.lsfb_isol.base import LSFBIsolBase


def _load_landmarks(
        root: str,
        videos: pd.DataFrame,
        landmarks: List[str],
        sequence_max_length: int,
        show_progress: bool,
):
    landmark_list = ', '.join(landmarks)
    print(f'Loading features ({landmark_list}) and labels for each isolated sign...')

    features = []
    targets = []

    progress_bar = tqdm(videos.iterrows(), total=videos.shape[0], disable=(not show_progress))
    for index, video in progress_bar:
        data = []

        for lm_type in landmarks:
            if lm_type == 'pose':
                data.append(load_pose_landmarks(root, video['pose']))
            elif lm_type == 'hand_left':
                data.append(load_hands_landmarks(root, video['hands'], 'left'))
            elif lm_type == 'hand_right':
                data.append(load_hands_landmarks(root, video['hands'], 'right'))
            else:
                raise ValueError(f'Unknown landmarks: {lm_type}.')

        features.append(pd.concat(data, axis=1).values[:sequence_max_length])
        targets.append(video['class'])

    return features, targets


class LSFBIsolLandmarks(LSFBIsolBase):

    def __init__(self, *args, **kwargs):
        print('-' * 10, 'LSFB ISOL DATASET')
        start_time = datetime.now()

        super(LSFBIsolLandmarks, self).__init__(*args, **kwargs)

        self.features, self.targets = _load_landmarks(
            self.config.root,
            self.config.videos,
            self.config.landmarks,
            self.config.sequence_max_length,
            self.config.show_progress,
        )
        self.labels = self.config.lemmes['lemme']

        print('-' * 10)
        print('loading time:', datetime.now() - start_time)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, index):
        features = self.features[index]
        target = self.targets[index]
        pad_value = 0

        if self.config.padding:
            pad_value = self.config.sequence_max_length - len(features)
            features = pad_landmarks(features, pad_value)

        if self.config.features_transform is not None:
            features = self.config.features_transform(features)

        if self.config.target_transform is not None:
            target = self.config.target_transform(target)

        if self.config.transform is not None:
            features, target = self.config.transform(features, target)

        if self.config.return_mask:
            mask = create_mask(self.config.sequence_max_length, pad_value, self.config.mask_value)
            if self.config.mask_transform is not None:
                mask = self.config.mask_transform(mask)

            return features, target, mask

        return features, target



import pandas as pd
import glob
from tqdm import tqdm
from os.path import sep
from concurrent.futures.process import ProcessPoolExecutor


def load_lsfb_dataset(path: str, verbose: bool = False):
    """
    Read the LSFB corpus in a dataframe containing the label of each signs, a number associated to each label,
    the path of the sign and which subset the signs come from (test or train).

    PARAMETERS:
      path : The path to lsfb corpus folder

    OUTPUT:
      dataset : A dataframe containing 4 columns (label, label_nbr, path, subset)
    """

    signs_folders = glob.glob(f"{path}{sep}*")
    map_label = {value.split(sep)[-1]: idx for (idx, value) in enumerate(signs_folders)}

    dataset = pd.DataFrame(columns=["label", "label_nbr", "path", "subset"])

    results = []
    with ProcessPoolExecutor() as executor:
        for folder in signs_folders:
            results.append(executor.submit(process_sign_folder, folder, map_label))

    df_data = []
    for idx, data in tqdm(enumerate(results), total=len(results), disable=not verbose):
        df_data += data.result()

    return pd.DataFrame(df_data)


def process_sign_folder(folder: str, map_label):
    train_path = f"{folder}{sep}train{sep}*"
    test_path = f"{folder}{sep}test{sep}*"
    label = folder.split(sep)[-1]

    data = []

    for sign in glob.glob(train_path):
        row = {}
        row["label"] = label
        row["label_nbr"] = map_label[label]
        row["path"] = sign
        row["subset"] = "train"

        data.append(row)

    for sign in glob.glob(test_path):
        row = {}
        row["label"] = label
        row["label_nbr"] = map_label[label]
        row["path"] = sign
        row["subset"] = "test"

        data.append(row)

    return data



"""
# Visualisation

This module provides tools to visualise the sign language video of the LSFB datasets.


"""

from .landmarks import plot_landmarks
from .landmarks import plot_grid
from .landmarks import plot_pose
from .landmarks import plot_hand




def duration_to_str(duration):
    milli = duration % 1000
    seconds = (duration // 1000) % 60
    minutes = duration // 60000
    return f'{minutes}min {seconds:2}s {milli:3}ms'


def get_gaps(vec, no_gap=0, gap=1):
    last = 0
    gaps = []

    last_state = 0
    for idx in range(len(vec)):
        val = vec[idx]

        if val == no_gap:
            if last_state == gap:
                gaps.append((last+1, idx-1))
            last = idx

        last_state = val

    return gaps



import cv2
import numpy as np
import pandas as pd


def absolute_position(img, pos):
    w = int(img.shape[1])
    h = int(img.shape[0])
    return int(max(0, min(w, w * pos[0]))), int(max(0, min(h, h * pos[1])))


def compute_box(pos, box_size):
    x, y = pos
    x1 = int(max(0, x - box_size / 2))
    y1 = int(max(0, y - box_size / 2))
    x2 = int(x + box_size / 2)
    y2 = int(y + box_size / 2)
    return x1, y1, x2, y2


def extract_box_img(img, box, dim=(256, 256)):
    x1, y1, x2, y2 = box
    box_img = img[y1:y2, x1:x2]

    if box_img.shape[0] != box_img.shape[1]:
        box_img = squarify_image(box_img)

    return cv2.resize(box_img, dim, interpolation=cv2.INTER_AREA)


def squarify_image(img):
    height = img.shape[0]
    width = img.shape[1]

    if height > width:
        padding = ((0, 0), (0, height-width), (0, 0))
    else:
        padding = ((0, width-height), (0, 0), (0, 0))

    return np.pad(img, padding, mode='constant', constant_values=0)


def get_feature_box(img, rel_pos: (float, float), holistic_features: pd.DataFrame, shoulder_length_factor: float):
    abs_pos = absolute_position(img, rel_pos)

    right_shoulder = np.array(absolute_position(img, (
        holistic_features['RIGHT_SHOULDER_X'],
        holistic_features['RIGHT_SHOULDER_Y']
    )))
    left_shoulder = np.array(absolute_position(img, (
        holistic_features['LEFT_SHOULDER_X'],
        holistic_features['LEFT_SHOULDER_Y']
    )))

    base_length = max(50, np.sqrt(np.sum(np.square(right_shoulder - left_shoulder))))
    return compute_box(abs_pos, base_length * shoulder_length_factor)



"""
# Module Datasets

This module provides configurations classes and iterators for the lsfb_iso and lsfb_cont dataset.
The module is organised in two submodules : One for the continuous dataset and one for the isolated one.

"""

from .lsfb_cont import LSFBContConfig, LSFBContLandmarks, LSFBContLandmarksGenerator
from .lsfb_isol import LSFBIsolConfig, LSFBIsolLandmarks, LSFBIsolLandmarksGenerator



"""
## lsfb_cont

This module contains classes helping to manipulate the lsfb_cont dataset. The provided classes are :

- **LSFBContConfig** : [dataclass](https://docs.python.org/3/library/dataclasses.html) allowing to configure how the dataset will be loaded
- **LSFBContLandmarks** : Iterator class loading the lsfb_cont landmarks according to the configuration provided.
- **LSFBContLandmarksGenerator** : Generator class loading the lsfb_cont landmarks according to the configuration provided.

"""

from .landmarks import LSFBContLandmarks
from .landmarks_generator import LSFBContLandmarksGenerator
from .config import LSFBContConfig



import abc
import numpy as np
import pandas as pd
import pickle
from os import path
from .config import LSFBContConfig
from ...utils.datasets import mini_sample, split_cont
from ...utils.target import (
    combine_binary_vectors,
    combine_binary_vectors_with_coarticulation,
)


class LSFBContBase:

    def __init__(self, config=None, **kwargs):
        print('-' * 10, 'LSFB CONT DATASET')

        self.config = LSFBContConfig(**kwargs) if config is None else config
        self.videos = self._load_video_list()

        self.targets: list[np.ndarray] = []
        self.labels: list[int] = []
        self.label_frequencies: list[int] = []
        self._load_targets()

    def _load_video_list(self) -> pd.DataFrame:
        split = self.config.split
        videos: pd.DataFrame = pd.read_csv(self.config.video_list_file)
        train_videos, test_videos = split_cont(videos, signers_frac=0.6, seed=42)

        if split == 'mini_sample':
            videos = mini_sample(videos, num_samples=10, seed=42)
        elif split == 'train':
            videos = train_videos
        elif split == 'test':
            videos = test_videos
        else:
            raise ValueError(f'Unknown split: {split}.')
        return videos

    def _load_targets(self):
        target = self.config.target
        targets_dir = self.config.targets_dir
        hands = self.config.hands

        if target == 'signs':
            filename = 'binary.pck'
            self.labels = ['waiting', 'signing']
            self.label_frequencies = [0.632736, 0.367264]
        elif target == 'signs_and_transitions':
            filename = 'binary_with_coarticulation.pck'
            self.labels = ['waiting', 'signing', 'coarticulation']
            self.label_frequencies = [0.519183, 0.367264, 0.113553]
        elif target == 'activity':
            filename = 'activity.pck'
            self.labels = ['waiting', 'signing']
            self.label_frequencies = [0.519183, 0.480817]
        else:
            raise ValueError(f'Unknown target: {target}.')

        target_filepath = path.join(targets_dir, filename)
        with open(target_filepath, 'rb') as file:
            target_vectors: dict = pickle.load(file)

        for _, video in self.videos.iterrows():
            filename = video['filename']
            vec = target_vectors.get(filename)

            if hands == 'left':
                vec = vec[0]
            elif hands == 'right':
                vec = vec[1]
            elif hands == 'both':
                if target == 'signs_and_transitions':
                    vec = combine_binary_vectors_with_coarticulation(vec[0], vec[1])
                else:
                    vec = combine_binary_vectors(vec[0], vec[1])

            assert vec is not None, f'Target not found for video {filename}.'
            self.targets.append(vec)

        print('Target vectors loaded.')

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def __getitem__(self, index):
        pass



import numpy as np
import cv2
from os import path
import random
from typing import Optional
import pandas as pd
from .cv import draw_holistic_landmarks, draw_holistic_boxes, get_holistic_features_img,\
    get_annotations_img, draw_pose_landmarks, draw_hands_landmarks, draw_face_landmarks, draw_face_mesh,\
    get_segmentation_vector_img
from ..utils.helpers import duration_to_str


class VideoPlayer:
    def __init__(self, video_filepath: str):
        if not path.isfile(video_filepath):
            raise FileNotFoundError(f"Video file not found: {video_filepath}")

        self.video_path: str = video_filepath
        self.draw_landmarks: bool = True
        self.draw_boxes: bool = True
        self.draw_duration: bool = True

        self.holistic_features: Optional[pd.DataFrame] = None

        self.right_hand_annotations: Optional[pd.DataFrame] = None
        self.left_hand_annotations: Optional[pd.DataFrame] = None

        self.pose_features: Optional[pd.DataFrame] = None
        self.hands_features: Optional[pd.DataFrame] = None
        self.face_features: Optional[pd.DataFrame] = None

        self.segmentations: list[tuple[str, np.ndarray]] = []

        self.face_meshing = False

        self.only_landmarks = False
        self.only_original = False
        self.crop = None

        self.pause = False
        self.speed = 1.0

    def change_speed(self, speed: float):
        assert 0 < speed, f'Invalid speed value ({speed}).'
        self.speed = speed

    def attach_holistic_features(self, holistic_filepath: str):
        if not path.isfile(holistic_filepath):
            raise FileNotFoundError(
                f"Holistic features file not found: {holistic_filepath}"
            )
        self.holistic_features = pd.read_csv(holistic_filepath)

    def attach_annotations(self, annotations_filepath: str, hand: str):
        if not path.isfile(annotations_filepath):
            raise FileNotFoundError(
                f"Annotations (CSV) file not found: {annotations_filepath}"
            )

        if hand == "right":
            self.right_hand_annotations = pd.read_csv(annotations_filepath)
        elif hand == "left":
            self.left_hand_annotations = pd.read_csv(annotations_filepath)
        else:
            raise ValueError(f"Unknown hand: {hand}")

    def attach_pose_landmarks(self, pose_filepath: str):
        if not path.isfile(pose_filepath):
            raise FileNotFoundError(f"Pose (CSV) file not found: {pose_filepath}")
        self.pose_features = pd.read_csv(pose_filepath)

    def attach_hands_landmarks(self, hands_filepath: str):
        if not path.isfile(hands_filepath):
            raise FileNotFoundError(f"Hands (CSV) file not found: {hands_filepath}")
        self.hands_features = pd.read_csv(hands_filepath)

    def attach_face_landmarks(self, face_landmarks: str):
        if not path.isfile(face_landmarks):
            raise FileNotFoundError(f"Face (CSV) file not found: {face_landmarks}")
        self.face_features = pd.read_csv(face_landmarks)

    def show_landmarks(self, value: bool):
        self.draw_landmarks = value

    def show_boxes(self, value: bool):
        self.draw_boxes = value

    def show_duration(self, value: bool):
        self.draw_duration = value

    def isolate_landmarks(self, value: bool):
        self.only_landmarks = value

    def isolate_original(self, value: bool):
        self.only_original = value

    def play(self):
        cap = cv2.VideoCapture(self.video_path)
        frame_rate = cap.get(cv2.CAP_PROP_FPS)
        frame_count = cap.get(cv2.CAP_PROP_FRAME_COUNT)

        frame_nb = 0
        frame_duration = int(1000 / frame_rate)
        delay = int(frame_duration / self.speed)
        total_duration = int(frame_count * 1000 / frame_rate)
        current_time = 0

        while frame_nb < frame_count and cap.isOpened():
            success, frame = cap.read()

            if not success:
                break

            key = cv2.waitKeyEx(delay)
            if key == ord('q'):
                break
            elif key == 32:
                self.pause = True
            elif key == 2424832:
                frame_nb = max(0, frame_nb - 50)
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_nb)
            elif key == 2555904:
                frame_nb = min(int(frame_count), frame_nb + 50)
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_nb)

            while self.pause:
                key = cv2.waitKeyEx(0)
                if key == 32:
                    self.pause = False

            self._show_frame(frame, frame_nb, current_time, total_duration)

            frame_nb += 1
            current_time += frame_duration

        cv2.destroyAllWindows()
        cap.release()

    def get_random_frame(self):
        cap = cv2.VideoCapture(self.video_path)
        frame_count = cap.get(cv2.CAP_PROP_FRAME_COUNT)
        frame_nb = random.randint(0, frame_count-1)

        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_nb)
        success, frame = cap.read()
        if not success:
            raise ValueError('Cannot read the frame.')
        cap.release()

        return cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    def _show_frame(self, frame, current_frame, current_time, total_duration):
        landmarks_frame = None
        original_frame = None

        if self.only_landmarks:
            landmarks_frame = np.zeros(frame.shape, dtype='uint8')
        if self.only_original:
            original_frame = frame.copy()

        self._draw_landmarks(frame, current_frame)
        if landmarks_frame is not None:
            self._draw_landmarks(landmarks_frame, current_frame)

        if self.holistic_features is not None:
            cv2.imshow('Features', get_holistic_features_img(frame, self.holistic_features.iloc[current_frame]))
            if self.draw_boxes:
                draw_holistic_boxes(frame, self.holistic_features.iloc[current_frame])

        if self.right_hand_annotations is not None:
            cv2.imshow(
                "Right hand annotations",
                get_annotations_img(self.right_hand_annotations, current_time),
            )
        if self.left_hand_annotations is not None:
            cv2.imshow(
                "Left hand annotations",
                get_annotations_img(self.left_hand_annotations, current_time),
            )

        for segmentation_name, segmentation in self.segmentations:
            cv2.imshow(
                f"Segmentation - {segmentation_name}",
                get_segmentation_vector_img(segmentation, current_time),
            )

        if landmarks_frame is not None:
            landmarks_frame = self._crop_frame(landmarks_frame)
            cv2.imshow('Landmarks', landmarks_frame)
        if original_frame is not None:
            original_frame = self._crop_frame(original_frame)
            cv2.imshow('Original', original_frame)

        frame = self._crop_frame(frame)
        if self.draw_duration:
            self._show_duration(frame, current_time, total_duration)
        cv2.imshow('Video', frame)
        cv2.setWindowProperty('Video', cv2.WND_PROP_TOPMOST, 1)

    def _draw_landmarks(self, frame, current_frame):
        if self.pose_features is not None:
            draw_pose_landmarks(frame, self.pose_features.iloc[current_frame].values.reshape((-1, 2)))
        if self.hands_features is not None:
            draw_hands_landmarks(frame, self.hands_features.iloc[current_frame].values.reshape((-1, 2)))
        if self.face_features is not None:
            if self.face_meshing:
                draw_face_mesh(frame, self.face_features.iloc[current_frame].values.reshape((-1, 2)))
            else:
                draw_face_landmarks(frame, self.face_features.iloc[current_frame].values.reshape((-1, 2)))
        if self.holistic_features is not None:
            draw_holistic_landmarks(frame, self.holistic_features.iloc[current_frame])

    def _crop_frame(self, frame):
        if self.crop is None:
            return frame

        x1, y1, x2, y2 = self.crop
        return frame[y1:y2, x1:x2]

    def _show_duration(self, frame, current_time: int, total_duration: int):
        duration_info = f'{duration_to_str(current_time)} / {duration_to_str(total_duration)}'
        cv2.putText(frame, duration_info, (20, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.6, (0, 255, 0), 1, 2)



from lsfb_dataset.utils.datasets import split_isol, mini_sample
from lsfb_dataset.datasets.lsfb_isol.config import LSFBIsolConfig
import abc


class LSFBIsolBase:
    """
    LSFB_ISOL Base Dataset.

    For each clips of LSFB_ISOL, this dataset provides a per-frame 2D-landmarks (skeleton)
    associated with a label for the performed sign (gloss).

    For each instance (video):
    **Features** are of size (F, L, 2) where
        - F is the number of frames
        - L the number of landmarks
    **Target** is a string representing the name of the gloss performed.

    Args:
        config: The configuration object (see LSFBContConfig).
            If config is not specified, every needed configuration argument must be manually provided.

    Configuration args (see LSFBIsolConfig):
        root: Root directory of the LSFB_ISOL dataset.
            The dataset must already be downloaded.
        landmarks: Select which landmarks (features) to use. Default = ['pose', 'hand_left', 'hand_right'].
            'pose' for pose skeleton (23 landmarks);
            'hands_left' for left hand skeleton (21 landmarks);
            'hands_right' for right hand skeleton (21 landmarks);
        features_transform: Callable object used to transform the features.
        target_transform: Callable object used to transform the targets.
        transform: Callable object used to transform both the features and the targets.
        mask_transform: Callable object used to transform the masks.
            You need to set return_mask to true to use this transform.
        lemmes_nb: Number of lemme to consider. Default=10
        lemme_list_path: Path to the csv containing the lemmes lists. Default="lemmes.csv"
        videos_list_path: Path to the csv containing the video information. Default="clips.csv"
        split: Select a specific subset of the dataset. Default = 'all'.
            'train' for training set;
            'test' for the test set;
            'all' for all the instances of the dataset;
            'mini_sample' for a tiny set of instances.
        sequence_max_length: Max length of the clip sequence. Default=50.
        padding: Pad all sequence to the same length.
        return_mask: Returning padding mask for the sequence.
        mask_value: Value of the masked part of the clips.
        show_progress: If true, show a progress bar while the dataset is loading.

    """

    def __init__(self, config=None, **kwargs):

        self.config = LSFBIsolConfig(**kwargs) if config is None else config
        self.config.videos = self._select_videos()

    def _select_videos(self):
        split = self.config.split
        videos = self.config.videos

        videos.drop(index=videos.index[~videos['class'].isin(self.config.lemmes.index)], inplace=True)

        if split == 'mini_sample':
            videos = mini_sample(videos)
        elif split != 'all':
            train_videos, test_videos = split_isol(videos)
            if split == 'train':
                videos = train_videos
            elif split == 'test':
                videos = test_videos

        return videos

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def __getitem__(self, index):
        pass



import pandas as pd


def get_annotations_in_time_range(df_annotations: pd.DataFrame, time_range: (int, int)):
    return df_annotations[
        ((df_annotations['start'] >= time_range[0]) & (df_annotations['start'] <= time_range[1])) |
        ((df_annotations['end'] >= time_range[0]) & (df_annotations['end'] <= time_range[1]))
    ]



from typing import Literal, NewType


DataSubset = NewType('DataSubset', Literal['all', 'train', 'test', 'mini_sample'])
Hand = NewType('Hand', Literal['left', 'right', 'both'])
Target = NewType('Target', Literal['activity', 'signs', 'signs_and_transitions'])



import matplotlib.pyplot as plt
from matplotlib.patches import Circle
from matplotlib.lines import Line2D
from typing import Optional
import numpy as np

from ..datasets.landmark_connections import POSE_CONNECTIONS, HAND_CONNECTIONS


def __compute_refocus(landmarks, focus_pad):
    """
    Compute the bounds of the frame that contains all the landmarks with a padding.

    Parameters
    ----------
    landmarks (Tensor): Features at an instant t that contains a flat representation
        of the landmarks that are contained in the refocused frame.
    focus_pad (float): The padding used for the refocus.

    Returns
    -------
    x_lim (float, float), y_lim (float, float): Boundaries of the refocused frame
    """
    landmarks = np.reshape(landmarks, (-1, 2))
    x = landmarks[:, 0]
    y = landmarks[:, 1]

    x_min = np.min(x)
    x_max = np.max(x)

    y_min = np.min(y)
    y_max = np.max(y)

    x_lim = (x_min - focus_pad, x_max + focus_pad)
    y_lim = (y_max + focus_pad, y_min - focus_pad)

    return x_lim, y_lim


def __compute_xy_lim(
        landmarks,
        aspect_ratio: float,
        *,
        x_lim: Optional[float] = None,
        y_lim: Optional[float] = None,
        refocus: bool = False,
        focus_pad: float = 0.02,
):
    if refocus:
        refocus_x, refocus_y = __compute_refocus(landmarks, focus_pad)
    else:
        refocus_x, refocus_y = (0, 1), (1, 0)

    if x_lim is None:
        x_lim = refocus_x
    if y_lim is None:
        y_lim = refocus_y

    x_size = x_lim[1] - x_lim[0]
    y_size = y_lim[0] - y_lim[1]
    aspect_ratio = (x_size * aspect_ratio / y_size)

    return x_lim, y_lim, aspect_ratio


def __draw_connections(x, y, *, ax, connections, edge_color):
    for v0, v1 in connections:
        line = Line2D(
            [x[v0], x[v1]],
            [y[v0], y[v1]],
            color=edge_color,
            zorder=1,
        )
        ax.add_line(line)


def __draw_vertices(landmarks, *, ax, vertex_size, vertex_color):
    for x, y in landmarks:
        ax.add_patch(Circle(
            (x, y),
            radius=vertex_size/2,
            facecolor=vertex_color,
            zorder=2,
        ))


def __plot_landmarks(
        landmarks,
        connections,
        *,
        ax=None,
        vertex_size: float = 0.01,
        vertex_color='lime',
        edge_color='white',
        background_color='black',
        size: int = 4,
        aspect_ratio=16/9,
        show_axis: bool = False,
        x_lim: Optional[tuple[float, float]] = None,
        y_lim: Optional[tuple[float, float]] = None,
        refocus: bool = False,
        focus_pad: float = 0.02,
):

    landmarks = np.reshape(landmarks, (-1, 2))
    x = landmarks[:, 0]
    y = landmarks[:, 1]

    x_lim, y_lim, aspect_ratio = __compute_xy_lim(
        landmarks,
        aspect_ratio=aspect_ratio,
        x_lim=x_lim,
        y_lim=y_lim,
        refocus=refocus,
        focus_pad=focus_pad,
    )

    fig = None
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(size * aspect_ratio, size))

    ax.set_facecolor(background_color)
    ax.axes.xaxis.set_visible(show_axis)
    ax.axes.yaxis.set_visible(show_axis)
    ax.set_xlim(*x_lim)
    ax.set_ylim(*y_lim)
    ax.set_box_aspect(1/aspect_ratio)

    __draw_vertices(landmarks, ax=ax, vertex_size=vertex_size, vertex_color=vertex_color)
    __draw_connections(x, y, ax=ax, connections=connections, edge_color=edge_color)

    if fig is not None:
        return fig, ax

    return ax


def plot_landmarks(
        landmarks,
        landmark_types: list[str],
        size: int = 4,
        aspect_ratio: float = 16/9,
        refocus: bool = False,
        focus_pad: float = 0.02,
        x_lim: Optional[tuple[float, float]] = None,
        y_lim: Optional[tuple[float, float]] = None,
        ax=None,
        **kwargs
):
    x_lim, y_lim, aspect_ratio = __compute_xy_lim(
        landmarks,
        aspect_ratio=aspect_ratio,
        x_lim=x_lim,
        y_lim=y_lim,
        refocus=refocus,
        focus_pad=focus_pad,
    )

    fig = None
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(size * aspect_ratio, size))

    last_index = 0
    for lm_type in landmark_types:
        if lm_type == 'pose':
            width = 46
            connections = POSE_CONNECTIONS
            vertex_color = 'lime'
            edge_color = 'white'
        elif lm_type == 'hand_left':
            width = 42
            connections = HAND_CONNECTIONS
            vertex_color = 'red'
            edge_color = 'red'
        elif lm_type == 'hand_right':
            width = 42
            connections = HAND_CONNECTIONS
            vertex_color = 'blue'
            edge_color = 'blue'
        else:
            raise ValueError(f'Unknown landmark type [{lm_type}].')

        __plot_landmarks(
            landmarks[last_index:last_index+width],
            connections,
            vertex_color=vertex_color,
            edge_color=edge_color,
            refocus=False,
            x_lim=x_lim,
            y_lim=y_lim,
            ax=ax,
            **kwargs
        )
        last_index += width

    if fig is not None:
        return fig, ax

    return ax


def plot_grid(
        landmarks,
        landmark_types: list[str],
        grid_size: tuple[int, int],
        indices=None,
        box_ratio: Optional[float] = None,
        figsize=(16, 9),
        **kwargs
):
    n_rows, n_cols = grid_size
    fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize, constrained_layout=True)

    if indices is None:
        indices = np.random.randint(landmarks.shape[0], size=(n_rows, n_cols))
        indices = np.sort(indices, axis=None).reshape(n_rows, n_cols)
    elif isinstance(indices, list):
        indices = np.array(indices)

    for row in range(n_rows):
        for col in range(n_cols):
            ax = plot_landmarks(
                landmarks[indices[row, col]],
                landmark_types,
                ax=axes[row, col],
                **kwargs
            )
            if box_ratio is not None:
                ax.set_box_aspect(box_ratio)

    return fig, axes


def plot_pose(pose_landmarks, **kwargs):
    __plot_landmarks(pose_landmarks, POSE_CONNECTIONS, **kwargs)


def plot_hand(hand_landmarks, **kwargs):
    __plot_landmarks(hand_landmarks, HAND_CONNECTIONS, **kwargs)



import numpy as np
import cv2
from lsfb_dataset.visualisation.cv_shapes import draw_connections, draw_points, draw_line, draw_rect
from lsfb_dataset.utils.annotations import get_annotations_in_time_range
from lsfb_dataset.utils.cv import get_feature_box, absolute_position, extract_box_img
from lsfb_dataset.datasets.landmark_connections import POSE_CONNECTIONS, HAND_CONNECTIONS,\
    FACEMESH_TESSELATION, FACEMESH_CONTOURS
from lsfb_dataset.utils.target import get_segments


def draw_feature_box(img, box: (int, int, int, int), color=(0, 255, 0), thickness=1):
    cv2.rectangle(img, (box[0], box[1]), (box[2], box[3]), color=color, thickness=thickness)


def draw_holistic_landmarks(img, features):
    face_pos = (features['FACE_X'], features['FACE_Y'])

    right_hand_pos = (features['RIGHT_HAND_X'], features['RIGHT_HAND_Y'])
    left_hand_pos = (features['LEFT_HAND_X'], features['LEFT_HAND_Y'])

    right_shoulder_pos = (features['RIGHT_SHOULDER_X'], features['RIGHT_SHOULDER_Y'])
    left_shoulder_pos = (features['LEFT_SHOULDER_X'], features['LEFT_SHOULDER_Y'])

    draw_line(img, right_shoulder_pos, left_shoulder_pos)
    draw_points(img, [right_hand_pos, right_shoulder_pos], color=(0, 0, 255))
    draw_points(img, [left_hand_pos, left_shoulder_pos], color=(255, 0, 0))
    draw_points(img, [face_pos], color=(0, 255, 0))


def draw_holistic_boxes(img, features):
    face_pos = (features['FACE_X'], features['FACE_Y'])
    right_hand_pos = (features['RIGHT_HAND_X'], features['RIGHT_HAND_Y'])
    left_hand_pos = (features['LEFT_HAND_X'], features['LEFT_HAND_Y'])

    draw_feature_box(img, get_feature_box(img, right_hand_pos, features, 0.8))
    draw_feature_box(img, get_feature_box(img, left_hand_pos, features, 0.8))
    draw_feature_box(img, get_feature_box(img, face_pos, features, 0.9), color=(0, 0, 255))


def get_holistic_features_img(img, features):
    info_img = np.zeros((256, 3 * 256, 3), dtype='uint8')

    face = features[['FACE_X', 'FACE_Y']].values
    right_hand = features[['RIGHT_HAND_X', 'RIGHT_HAND_Y']].values
    left_hand = features[['LEFT_HAND_X', 'LEFT_HAND_Y']].values
    right_shoulder = features[['RIGHT_SHOULDER_X', 'RIGHT_SHOULDER_Y']].values
    left_shoulder = features[['LEFT_SHOULDER_X', 'LEFT_SHOULDER_Y']].values

    if np.isnan(right_shoulder).sum() or np.isnan(left_shoulder).sum():
        return info_img

    if not np.isnan(right_hand.sum()):
        info_img[:, :256] = extract_box_img(img, get_feature_box(img, right_hand, features, 0.8))

    if not np.isnan(left_hand.sum()):
        info_img[:, 256:512] = extract_box_img(img, get_feature_box(img, left_hand, features, 0.8))

    if not np.isnan(face.sum()):
        info_img[:, 512:] = extract_box_img(img, get_feature_box(img, face, features, 0.9))

    return info_img


def draw_annot_rect(img, x_start, x_end):
    draw_rect(img, (x_start, 0.5), (x_end, 1.0), color=(51, 204, 51))


def draw_annot_text(img, text, x_start, y_text):
    draw_line(img, (x_start, 1), (x_start, y_text - 0.05), color=(51, 204, 51), thickness=1)
    cv2.putText(img, text, absolute_position(img, (x_start, y_text)),
                cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1, 2)


def draw_annotation(img, annot, time, index=0):
    x_start = (annot['start'] - time + 3000) / 6000
    x_end = (annot['end'] - time + 3000) / 6000
    draw_annot_rect(img, x_start, x_end)

    y_text = 0.4 - (index % 5) * 0.07
    draw_annot_text(img, annot['gloss'], x_start, y_text)


def get_annotations_img(df_annot, time, time_offset=3000, width=512, height=256):
    annot_img = np.zeros((height, width, 3), dtype='uint8')

    df_current_annotations = get_annotations_in_time_range(df_annot, (time - time_offset, time + time_offset))
    for index, annot in df_current_annotations.iterrows():
        draw_annotation(annot_img, annot, time, index=index)
    draw_line(annot_img, (0.5, 0), (0.5, 1.0))

    return annot_img


def draw_pose_landmarks(img, positions):
    draw_connections(img, positions, POSE_CONNECTIONS, color=(0, 255, 0))
    draw_points(img, positions)


def draw_hands_landmarks(img, positions):
    draw_hand_landmarks(img, positions[:21], color=(255, 0, 0))
    draw_hand_landmarks(img, positions[21:])


def draw_hand_landmarks(img, positions, color=(0, 0, 255)):
    draw_connections(img, positions, HAND_CONNECTIONS, color)


def draw_face_landmarks(img, positions, color=(0, 255, 0)):
    draw_connections(img, positions, FACEMESH_CONTOURS, color, thickness=1)
    draw_points(img, positions, color=(0, 0, 255), radius=1)


def draw_face_mesh(img, positions, color=(0, 255, 0)):
    draw_connections(img, positions, FACEMESH_TESSELATION, color, thickness=1)


def draw_indices(img, positions, color=(0, 0, 255), font_scale=1, thickness=2, offset=(10, 10)):
    for index, pos in enumerate(positions):
        x, y = absolute_position(img, pos)
        x += offset[0]
        y += offset[1]
        cv2.putText(img, str(index), (x, y), cv2.FONT_HERSHEY_SIMPLEX, font_scale, color, thickness, cv2.LINE_AA)


def draw_segment_rect(img, x_start, x_end):
    draw_rect(img, (x_start, 0.5), (x_end, 1.0), color=(66, 150, 235))


def get_segmentation_vector_img(segmentation, time, width=512, height=256):
    seg_img = np.zeros((height, width, 3), dtype='uint8')

    bound_start = int((max(0, time - 3000))/20)
    bound_end = int((time + 3000)/20)

    segmentation = segmentation[bound_start:bound_end]
    segments = get_segments(segmentation, filter_value=1) + bound_start

    for frame_start, frame_end in segments:
        time_start = 20 * frame_start
        time_end = 20 * frame_end

        x_start = (time_start - time + 3000) / 6000
        x_end = (time_end - time + 3000) / 6000

        draw_segment_rect(seg_img, x_start, x_end)

    draw_line(seg_img, (0.5, 0), (0.5, 1.0))

    return seg_img



import pandas as pd
import numpy as np


def split_cont(df_videos: pd.DataFrame, signers_frac=0.6, seed=42):
    signers = pd.Series(df_videos['signer'].unique())
    train_signers = signers.sample(frac=signers_frac, random_state=seed)
    val_signers = signers.drop(index=train_signers.index)
    train_df = df_videos[df_videos['signer'].isin(train_signers)]
    val_df = df_videos[df_videos['signer'].isin(val_signers)]
    return train_df, val_df


def split_isol(dataframe: pd.DataFrame, test_frac=0.25, seed=42):
    test_df = dataframe.sample(frac=test_frac, random_state=seed)
    train_df = dataframe.drop(index=test_df.index)
    return train_df, test_df


def mini_sample(dataframe: pd.DataFrame, num_samples: int = 10, seed=42):
    return dataframe.sample(n=num_samples, random_state=seed)


def create_mask(seq_len: int, padding: int, mask_value: int = 0):
    mask = np.ones(seq_len, dtype='uint8')
    mask[-padding:] = mask_value
    return mask



import numpy as np
from sklearn.preprocessing import MinMaxScaler


class MinMaxRescale(object):

    def __call__(self, features: np.ndarray):
        scaler = MinMaxScaler(copy=False)
        return scaler.fit_transform(features)



import torch


class ToTensor(object):

    def __init__(self, dtype=None):
        self.dtype = dtype

    def __call__(self, array):
        array = torch.from_numpy(array)

        if self.dtype is not None:
            array = array.type(self.dtype)

        return array



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lsfb-dataset",
    version="0.1.0",
    author="Jérôme Fink, Poitier Pierre",
    author_email="jerome.fink@unamur.be, pierre.poitier@unamur.be",
    description="A companion library for the LSFB-dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jefidev/lsfb-dataset",
    project_urls={
        "Bug Tracker": "https://github.com/Jefidev/lsfb-dataset/issues",
        "Documentation": "https://jefidev.github.io/lsfb-dataset/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas",
        "opencv-python",
        "tqdm",
    ],
)



from tqdm import tqdm
import numpy as np
from .base import LSFBContBase
from ...utils.datasets import create_mask
from ...utils.landmarks import (
    load_landmarks,
    pad_landmarks,
)
from ...utils.target import pad_target


class LSFBContLandmarks(LSFBContBase):
    """
    Utility class to load the LSFB CONT Landmarks dataset.
    The dataset must be already downloaded!

    All the landmarks and targets are loaded into memory (RAM) to maximise the efficiency of the training.
    If you don't have enough RAM (between 8GB and 16GB) and don't want to load all of them in memory,
    use the LSFBContLandmarksGenerator class instead.

    Properties:
        features: The list of landmarks for each instance of the dataset
            For each instance i, (L,D) were
                L is the length of the sequence
                D is the number of features depending on the landmarks selected in the configuration
        targets: The list of target segmentations for each instance of the dataset

        labels: The targeted labels in the dataset. Example: waiting, signing and coarticulation.
            The labels depend on the targeted segmentation.
        label_frequencies: The frequency of each label used in the dataset.
            The labels depend on the targeted segmentation.

        windows: List of every window in the dataset
            (instance_index, window_start, window_end, padding)
            This list only contains windows if the `window` configuration is set!

    Args:
        config: The configuration object (see LSFBContConfig).
            If config is not specified, every needed configuration argument must be manually provided.

    Author: ppoitier
    """

    def __init__(self, **kwargs):
        super(LSFBContLandmarks, self).__init__(**kwargs)

        self.features: list[np.ndarray] = []
        self._load_features()

        # (video_index, window_start, window_end, padding)
        self.windows: list[tuple[int, int, int, int]] = []
        if self.config.window is not None:
            self._build_windows()

    def __len__(self):
        if self.config.window is not None:
            return len(self.windows)
        return len(self.features)

    def __getitem__(self, index):
        features_transform = self.config.features_transform
        target_transform = self.config.target_transform
        transform = self.config.transform

        window = self.config.window
        return_mask = self.config.return_mask
        mask_transform = self.config.mask_transform

        if window is not None:
            landmarks, target, padding = self._get_windowed_item(index)
        else:
            landmarks = self.features[index]
            target = self.targets[index]
            padding = 0

        if features_transform is not None:
            landmarks = features_transform(landmarks)

        if target_transform is not None:
            target = target_transform(target)

        if transform is not None:
            landmarks, target = transform(landmarks, transform)

        if return_mask:
            mask = create_mask(len(target), padding)
            if mask_transform is not None:
                mask = mask_transform(mask)
            return landmarks, target, mask

        return landmarks, target

    def _load_features(self):
        root = self.config.root
        landmarks = self.config.landmarks
        show_progress = self.config.show_progress

        landmarks_nb = self.videos.shape[0]
        landmarks_list = ', '.join(landmarks)
        print(f'Loading landmarks {landmarks_list} for {landmarks_nb} videos...')

        progress_bar = tqdm(
            self.videos.iterrows(),
            total=landmarks_nb,
            disable=(not show_progress),
        )

        for _, video in progress_bar:
            self.features.append(load_landmarks(video, root, landmarks))
        print('Landmarks loaded.')

    def _build_windows(self):
        window = self.config.window
        window_size, window_stride = window
        for index, landmarks in enumerate(self.features):
            landmarks_nb = landmarks.shape[0]
            for start in range(0, landmarks_nb, window_stride):
                end = min(landmarks_nb, start + window_size)
                padding = window_size - (end - start)
                self.windows.append((index, start, end, padding))
        print('Windows successfully created.')

    def _get_windowed_item(self, index: int):
        video_index, start, end, padding = self.windows[index]
        landmarks = pad_landmarks(self.features[video_index][start:end], padding)
        target = pad_target(self.targets[video_index][start:end], padding)
        return landmarks, target, padding



import numpy as np
from itertools import groupby
from typing import Optional
from os import path
import pickle


def pad_target(target, padding: int):
    if padding > 0:
        target = np.pad(target, (0, padding), constant_values=(0, 0))
    return target


def combine_binary_vectors(vec1, vec2):
    return np.logical_or(vec1 == 1, vec2 == 1).astype('uint8')


def combine_binary_vectors_with_coarticulation(vec1, vec2):
    vec = 2 * np.logical_or(vec1 == 2, vec2 == 2).astype('uint8')
    vec[vec1 == 1] = 1
    vec[vec2 == 1] = 1
    return vec


def get_segments(segmentation: np.ndarray, filter_value: Optional[int] = None):
    segments = np.array([[key, 0, len(list(val))] for key, val in groupby(segmentation)])
    segments[1:, 1] = np.cumsum(segments[:-1, 2], axis=0)
    segments[:, 2] += segments[:, 1] - 1

    if filter_value is not None:
        segments = segments[segments[:, 0] == filter_value][:, 1:]

    return segments


def load_segmentation_target(
        root: str,
        video_filename: str,
        target_vectors_dir: str = 'annotations/vectors',
        target_type: str = 'signs',
):
    if target_type == 'signs':
        vectors_filename = 'binary.pck'
    elif target_type == 'signs_and_transitions':
        vectors_filename = 'binary_with_coarticulation.pck'
    elif target_type == 'activity':
        vectors_filename = 'activity.pck'
    else:
        raise ValueError(f'Unknown target: {target_type}')

    targets_filepath = path.join(root, target_vectors_dir, vectors_filename)
    if not path.isfile(targets_filepath):
        raise FileNotFoundError(f'Target vectors ({target_type}) file not found')

    with open(targets_filepath, 'rb') as file:
        vectors = pickle.load(file)

    target = vectors.get(video_filename)
    if target is None:
        raise ValueError(f'Target ({target_type}) not found for video [{video_filename}].')

    return target




import pandas as pd

from lsfb_dataset.utils.landmarks import load_pose_landmarks, load_hands_landmarks, pad_landmarks
from lsfb_dataset.utils.datasets import create_mask
from lsfb_dataset.datasets.lsfb_isol.base import LSFBIsolBase


class LSFBIsolLandmarksGenerator(LSFBIsolBase):

    def __init__(self, *args, **kwargs):
        super(LSFBIsolLandmarksGenerator, self).__init__(*args, **kwargs)
        self.labels = self.config.lemmes['lemme']

    def __len__(self):
        return len(self.config.videos)

    def __getitem__(self, index):

        video = self.config.videos.iloc[index]

        features, target = self._load_landmark(video)
        pad_value = 0

        if self.config.padding:
            pad_value = self.config.sequence_max_length - len(features)
            features = pad_landmarks(features, pad_value)

        if self.config.features_transform is not None:
            features = self.config.features_transform(features)

        if self.config.target_transform is not None:
            target = self.config.target_transform(target)

        if self.config.transform is not None:
            features, target = self.config.transform(features, target)

        if self.config.return_mask:
            mask = create_mask(self.config.sequence_max_length, pad_value, self.config.mask_value)
            if self.config.mask_transform is not None:
                mask = self.config.mask_transform(mask)

            return features, target, mask

        return features, target

    def _load_landmark(self, video):
        data = []

        landmarks = self.config.landmarks
        root = self.config.root

        for lm_type in landmarks:
            if lm_type == 'pose':
                data.append(load_pose_landmarks(root, video['pose']))
            elif lm_type == 'hand_left':
                data.append(load_hands_landmarks(root, video['hands'], 'left'))
            elif lm_type == 'hand_right':
                data.append(load_hands_landmarks(root, video['hands'], 'right'))
            else:
                raise ValueError(f'Unknown landmarks: {lm_type}.')

        features = pd.concat(data, axis=1).values[:self.config.sequence_max_length]
        targets = video['class']

        return features, targets



import pandas as pd
import numpy as np
from os import path
from typing import Optional


def load_pose_landmarks(root: Optional[str], csv_path: str) -> pd.DataFrame:
    filepath = path.join(root, csv_path) if root is not None else csv_path
    df = pd.read_csv(filepath, dtype='float32')
    return df


def load_hands_landmarks(root: Optional[str], csv_path: str, hands: str) -> pd.DataFrame:
    filepath = path.join(root, csv_path) if root is not None else csv_path
    df = pd.read_csv(filepath, dtype='float32')
    if hands == 'right':
        df = df.loc[:, df.columns.str.startswith('RIGHT')]
    elif hands == 'left':
        df = df.loc[:, df.columns.str.startswith('LEFT')]
    return df


def load_landmarks(video: pd.Series, root: str, landmarks: list[str]) -> np.ndarray:
    video_landmarks = []
    for landmark in landmarks:
        if landmark == 'pose':
            video_landmarks.append(load_pose_landmarks(root, video['pose']))
        elif landmark == 'hand_left':
            video_landmarks.append(load_hands_landmarks(root, video['hands'], 'left'))
        elif landmark == 'hand_right':
            video_landmarks.append(load_hands_landmarks(root, video['hands'], 'right'))
        else:
            raise ValueError(f'Unknown landmarks: {landmark}.')
    return pd.concat(video_landmarks, axis=1).values


def pad_landmarks(landmarks, padding: int):
    if padding > 0:
        landmarks = np.pad(landmarks, ((0, padding), (0, 0)), constant_values=((0, 0), (0, 0)))
    return landmarks



import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

from ..utils.helpers import get_gaps
from ..utils.target import get_segments


def plot_binary_segmentation(vector, x_lim: tuple[int, int] = None):
    if x_lim is None:
        x_lim = 0, len(vector)

    fig, ax = plt.subplots(1, 1, figsize=(30, 4))
    ax.set_xlim(*x_lim)

    for start, end in get_gaps(vector, 0, 1):
        plt.axvspan(start, end)

    return fig, ax


def plot_binary_segmentation_with_coarticulation(vector, x_lim: tuple[int, int] = None):
    if x_lim is None:
        x_lim = 0, len(vector)

    fig, ax = plt.subplots(1, 1, figsize=(30, 4))
    ax.set_xlim(*x_lim)

    for start, end in get_gaps(vector, 0, 1):
        plt.axvspan(start, end)

    for start, end in get_gaps(vector, 1, 2):
        plt.axvspan(start, end, color='red')


def plot_lemme_segmentation(
        segmentation: np.ndarray,
        x_lim: tuple[int, int] = None,
        lemmes_nb: int = 100,
        cmap: str = 'twilight_shifted',
):
    if x_lim is None:
        x_lim = 0, len(segmentation)

    norm = mpl.colors.Normalize(vmin=0, vmax=2+lemmes_nb)
    cmap = cm.get_cmap(cmap)

    fig, ax = plt.subplots(1, 1, figsize=(30, 4))
    ax.set_xlim(*x_lim)

    for value, start, end in get_segments(segmentation):
        if value == 0:
            continue
        plt.axvspan(start, end, color=cmap(norm(value)))



import numbers
import random
import numpy as np
import PIL

import lsfb_dataset.utils.video_utils as F


class I3DPixelsValue(object):
    """
    Scale the pixel value between -1 and 1 instead of 0 and 1 (required for I3D)
    """

    def __call__(self, sample):
        return sample * 2 - 1


class ChangeVideoShape(object):
    """
    Expect to receive a ndarray of chape (Time, Height, Width, Channel) which is the default format
    of cv2 or PIL. Change the shape of the ndarray to TCHW or CTHW.
    """

    def __init__(self, shape: str):
        """
        shape : a string with the value "CTHW" or "TCHW".
        """

        self.shape = shape

    def __call__(self, sample):

        if self.shape == "CTHW":
            sample = np.transpose(sample, (3, 0, 1, 2))
        elif self.shape == "TCHW":
            sample = np.transpose(sample, (0, 3, 1, 2))
        else:
            raise ValueError(f"Received {self.shape}. Expecting TCHW or CTHW.")

        return sample


class ResizeVideo(object):
    """Resizes a list of ( T x H x W x C) to the final size
    The larger the original image is, the more times it takes to
    interpolate
    Args:
    interpolation (str): Can be one of 'nearest', 'bilinear'
    defaults to nearest
    size (tuple): (widht, height)
    """

    def __init__(self, size, interpolation="nearest"):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, clip):
        resized = F.resize_clip(clip, self.size, interpolation=self.interpolation)
        return np.array(resized)


class RandomCropVideo(object):
    """Extract random crop at the same location for a list of images
    Args:
    size (sequence or int): Desired output size for the
    crop in format (h, w)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            size = (size, size)

        self.size = size

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray
        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        h, w = self.size
        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size
        else:
            raise TypeError(
                "Expected numpy.ndarray or PIL.Image"
                + "but got list of {0}".format(type(clip[0]))
            )
        if w > im_w or h > im_h:
            error_msg = (
                "Initial image size should be larger then "
                "cropped size but got cropped sizes : ({w}, {h}) while "
                "initial image is ({im_w}, {im_h})".format(
                    im_w=im_w, im_h=im_h, w=w, h=h
                )
            )
            raise ValueError(error_msg)

        x1 = random.randint(0, im_w - w)
        y1 = random.randint(0, im_h - h)
        cropped = F.crop_clip(clip, y1, x1, h, w)

        return np.array(cropped)


class CenterCropVideo(object):
    """Extract center crop at the same location for a list of images
    Args:
    size (sequence or int): Desired output size for the
    crop in format (h, w)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            size = (size, size)

        self.size = size

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray
        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        h, w = self.size
        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size
        else:
            raise TypeError(
                "Expected numpy.ndarray or PIL.Image"
                + "but got list of {0}".format(type(clip[0]))
            )
        if w > im_w or h > im_h:
            error_msg = (
                "Initial image size should be larger then "
                "cropped size but got cropped sizes : ({w}, {h}) while "
                "initial image is ({im_w}, {im_h})".format(
                    im_w=im_w, im_h=im_h, w=w, h=h
                )
            )
            raise ValueError(error_msg)

        x1 = int(round((im_w - w) / 2.0))
        y1 = int(round((im_h - h) / 2.0))
        cropped = F.crop_clip(clip, y1, x1, h, w)

        return np.array(cropped)


class TrimVideo(object):
    """Trim each video the same way. Waiting shape TCHW"""

    def __init__(self, size, offset=None):
        self.end = size
        self.begin = 0

        if offset is not None:
            self.begin = offset
            self.end += offset

    def __call__(self, clip):
        resized = clip

        if len(clip) > self.end:
            resized = clip[self.begin:self.end]
        return np.array(resized)


class RandomTrimVideo(object):
    """Trim randomly the video. Waiting shape TCHW"""

    def __init__(self, size):
        self.size = size

    def __call__(self, clip):
        resized = clip

        if len(clip) > self.size:
            diff = len(resized) - self.size

            start = random.randint(0, diff)
            end = start + self.size

            resized = resized[start:end]

        return np.array(resized)


class PadVideo(object):
    def __init__(self, size, loop=True):
        self.size = size
        self.loop = loop

    def __call__(self, clip):
        if self.loop:
            resized = self._loop_sequence(clip, self.size)
        else:
            resized = self._pad_sequence(clip, self.size)

        return np.array(resized)

    def _pad_sequence(self, sequence, length):
        shape = sequence.shape
        new_shape = (length, shape[1], shape[2], shape[3])

        zero_arr = np.zeros(new_shape)
        zero_arr[: shape[0]] = sequence

        return zero_arr

    def _loop_sequence(self, sequence, length):
        shape = sequence.shape
        new_shape = (length, shape[1], shape[2], shape[3])
        zero_arr = np.zeros(new_shape)

        video_len = len(sequence)

        for i in range(length):
            vid_idx = i % video_len
            zero_arr[i] = sequence[vid_idx]

        return zero_arr
